/*
 * cache.c
 * Copyright (C) 2010 - 2018 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

#if defined (YASAPI) // [
#if defined (CACHE_CRITICAL_SECTION) // [
#define CacheLock(pCache) \
    EnterCriticalSection(&(pCache)->cs)
#define CacheUnlock(pCache) \
    LeaveCriticalSection(&(pCache)->cs)
#else // ] [
#define CacheLock(pCache)
#define CacheUnlock(pCache)
#endif // ]

void CacheCreate(Cache *pCache)
{
#if defined (CACHE_CRITICAL_SECTION) // [
  InitializeCriticalSection(&pCache->cs);
#endif // ]
}

void CacheDestroy(Cache *pCache)
{
#if defined (CACHE_CRITICAL_SECTION) // [
  DeleteCriticalSection(&pCache->cs);
#endif // ]
}

void CacheSet(Cache *pCache, const char *samples, size_t nSizeFFSoX,
    size_t nSizeYasapi)
{
  CacheLock(pCache); // {
  pCache->size=nSizeFFSoX;
  pCache->step=nSizeYasapi<nSizeFFSoX?nSizeYasapi:nSizeFFSoX;
  pCache->rp=samples;
  pCache->mp=samples+nSizeFFSoX;
  CacheUnlock(pCache); // }
}

void CacheReset(Cache *pCache)
{
  CacheLock(pCache); // {
  pCache->size=0;
  pCache->step=0;
  pCache->rp=NULL;
  pCache->mp=NULL;
  CacheUnlock(pCache); // }
}

int CacheValid(Cache *pCache)
{
  int bValid;

  CacheLock(pCache); // {
  bValid=pCache->size&&pCache->rp<pCache->mp;
  CacheUnlock(pCache); // }

  return bValid;
}

size_t CacheSizeStep(Cache *pCache)
{
  size_t size;

  CacheLock(pCache); // {
  size=pCache->rp+pCache->step<pCache->mp?pCache->step:pCache->mp-pCache->rp;
  CacheUnlock(pCache); // }

  return size;
}

size_t CacheSizeLatency(Cache *pCache)
{
  size_t latency;

  CacheLock(pCache); // {
  latency=pCache->size?pCache->mp-pCache->rp:0u;
  CacheUnlock(pCache); // }

  return latency;
}

void CacheNext(Cache *pCache)
{
  CacheLock(pCache); // {
  pCache->rp+=CacheSizeStep(pCache);
  CacheUnlock(pCache); // }
}
#endif // ]
