/*
 * ffsox.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

#if ! defined (AVUTIL_VER) // [
#define AVUTIL_VER 54
#endif // ]
#if ! defined (SWRESAMPLE_VER) // [
#define SWRESAMPLE_VER 1
#endif // ]
#if ! defined (AVCODEC_VER) // [
#define AVCODEC_VER 56
#endif // ]
#if ! defined (AVFORMAT_VER) // [
#define AVFORMAT_VER 56
#endif // ]
#if ! defined (LIBSOX_VER) // [
#define LIBSOX_VER 2
#endif // ]

///////////////////////////////////////////////////////////////////////////////
static struct {
  CRITICAL_SECTION *buf;
  CRITICAL_SECTION *wp;
  CRITICAL_SECTION *mp;
} ffmpeg_cs;

static int open_ffmpeg_locks()
{
  size_t size=8;

  if (NULL==(ffmpeg_cs.buf=malloc(size*(sizeof *ffmpeg_cs.buf))))
    goto error;

  ffmpeg_cs.wp=ffmpeg_cs.buf;
  ffmpeg_cs.mp=ffmpeg_cs.buf+size;

  return 0;
error:
  return -1;
}

#if defined (CLEANUP) // {
static void close_ffmpeg_locks()
{
  if (NULL!=ffmpeg_cs.buf)
    free(ffmpeg_cs.buf);
}
#endif // }

//#define FFMPEG_LOCK_CALLBACK
#if defined (FFMPEG_LOCK_CALLBACK) // [
static int ffmpeg_lock_callback(void **mutex, enum AVLockOp op)
{
  switch (op) {
    case AV_LOCK_CREATE:
      if (ffmpeg_cs.wp==ffmpeg_cs.mp) {
        size_t size=ffmpeg_cs.mp-ffmpeg_cs.buf;

        if (NULL==(ffmpeg_cs.buf=realloc(ffmpeg_cs.buf,
            2*size*(sizeof *ffmpeg_cs.buf))))
          goto error;

        ffmpeg_cs.wp=ffmpeg_cs.buf+size;
        ffmpeg_cs.mp=ffmpeg_cs.buf+2*size;
      }

      InitializeCriticalSection(ffmpeg_cs.wp);
      *mutex=ffmpeg_cs.wp;
      ++ffmpeg_cs.wp;
      break;
    case AV_LOCK_OBTAIN:
      EnterCriticalSection(*mutex);
      break;
    case AV_LOCK_RELEASE:
      LeaveCriticalSection(*mutex);
      break;
    case AV_LOCK_DESTROY:
      DeleteCriticalSection(*mutex);
      break;
    default:
      goto error;
  }

  return 0;
error:
  return -1;
}
#endif // ]

///////////////////////////////////////////////////////////////////////////////
typedef struct ffsox_sym {
  pb_str_t *pfx;
  pb_str_t *bn;
  pb_str_t *ver;
} ffsox_sym_t;

#if defined (SOX) // [
static ffsox_sym_t ffsox_sox[]={
  { NULL,_PS("libsox"),_PS(STRA(LIBSOX_VER)) },
  { NULL,NULL,NULL }
};
#endif // ]

static ffsox_sym_t ffsox_ffmpeg[]={
  { _PS("lib"),_PS("avutil"),_PS(STRA(AVUTIL_VER)) },
  { _PS("lib"),_PS("swresample"),_PS(STRA(SWRESAMPLE_VER)) },
  { _PS("lib"),_PS("swscale"),_PS(STRA(SWSCALE_VER)) },
  { _PS("lib"),_PS("postproc"),_PS(STRA(POSTPROC_VER)) },
  { _PS("lib"),_PS("avcodec"),_PS(STRA(AVCODEC_VER)) },
  { _PS("lib"),_PS("avformat"),_PS(STRA(AVFORMAT_VER)) },
  { _PS("lib"),_PS("avfilter"),_PS(STRA(AVFILTER_VER)) },
  { NULL,NULL,NULL }
};

static ffsox_sym_t ffsox_dlgitemurl[]={
  { NULL,_PS("libdlgitemurl"),NULL },
  { NULL,NULL,NULL }
};

static int ffsox_load_sym(pb_str_t *home, ffsox_sym_t *e)
{
  int code;

  while (e->bn) {
    if (0!=(code=pb_load(home,e->pfx,e->bn,e->ver)))
      return code;

    ++e;
	}

  return 0;
}

#if defined (SOX) // [
int ffsox_load_sox(pb_str_t *home)
{
  if (0!=ffsox_load_sym(home,ffsox_sox)) {
    DWRITELN("Error loading SoX.");
    return 1;
  }
  else if (SOX_SUCCESS!=sox_init()) {
    DWRITELN("Error initializing SoX.");
    return 1;
  }
  else {
    DWRITELN("SoX sucessfully loaded.");
    return 0;
  }
}
#endif // ]

int ffsox_load_ffmpeg(pb_str_t *home)
{
  if (0!=ffsox_load_sym(home,ffsox_ffmpeg)) {
    DWRITELN("Error loading FFmpeg.");
    return -1;
  }

#if 0 // [
  av_register_all();
  avfilter_register_all();
#endif // ]
#if defined (DEBUG) // {
  av_log_set_level(AV_LOG_ERROR);
#else // } {
  av_log_set_level(AV_LOG_QUIET);
#endif // }

  if (0!=open_ffmpeg_locks()) {
    DWRITELN("Error creating FFmpeg locks.");
    return -1;
  }

#if defined (FFMPEG_LOCK_CALLBACK) // [
  if (0!=av_lockmgr_register(ffmpeg_lock_callback)) {
    DWRITELN("Error registering FFmpeg lock manager.");
    return -1;
  }
#endif // ]

  DWRITELN("FFmpeg sucessfully loaded.");

  return 0;
}

int ffsox_load_dlgitemurl(pb_str_t *home)
{
  if (0!=ffsox_load_sym(home,ffsox_dlgitemurl)) {
    DWRITELN("Error loading DlgItemURL.");
    return 1;
  }
  else {
    DWRITELN("DlgItemURL sucessfully loaded.");
    return 0;
  }
}

#if defined (CLEANUP) // {
void ffsox_unload_ffmpeg()
{
  av_lockmgr_register(NULL);
  close_ffmpeg_locks();
}
#endif // }
