/*
 * samples.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

///////////////////////////////////////////////////////////////////////////////
#define DEFAULT_GAIN      (-12.0)
#define DEFAULT_ADAPTIVE  5

///////////////////////////////////////////////////////////////////////////////
typedef uint8_t u8_t;
typedef int16_t s16_t;
typedef int32_t s32_t;
typedef float f32_t;
typedef double f64_t;

///////////////////////////////////////////////////////////////////////////////
samples_t *samples_init(samples_t *s)
{
  DMARKLN();

  s->hMutex=CreateMutex(
    NULL,   // _In_opt_  LPSECURITY_ATTRIBUTES lpMutexAttributes,
    FALSE,  // _In_      BOOL bInitialOwner,
    NULL    // _In_opt_  LPCTSTR lpName
  );

  if (NULL==s->hMutex)
    goto error;

  s->hEventWrite=CreateEvent(
    NULL,   // lpEventAttributes
    FALSE,  // bManualReset
    FALSE,  // bInitialState
    NULL    // lpName
  );

  if (NULL==s->hEventWrite)
    goto error;

  s->hEventRead=CreateEvent(
    NULL,   // lpEventAttributes
    FALSE,  // bManualReset
    FALSE,  // bInitialState
    NULL    // lpName
  );

  if (NULL==s->hEventRead)
    goto error;

#if defined (IN_FFSOX_BS1770) // {
  bs1770_lufs_init(&s->lufs,0);
#endif // }
  return s;
error:
  samples_cleanup(s);

  return NULL;
}

samples_t *samples_cleanup(samples_t *s)
{
#if defined (IN_FFSOX_BS1770) // {
  bs1770_lufs_cleanup(&s->lufs);
#endif // }

  if (NULL!=s->in.rep)
    free(s->in.rep);

  if (NULL!=s->hEventRead)
    CloseHandle(s->hEventRead);

  if (NULL!=s->hEventWrite)
    CloseHandle(s->hEventWrite);

  if (NULL!=s->hMutex)
    CloseHandle(s->hMutex);

  return s;
}

void samples_kill(samples_t *s)
{
  WaitForSingleObject(s->hMutex,INFINITE);

  if (STATE_RUN==s->state) {
    s->in.wp=s->in.rep;
    s->state=STATE_KILL;
    SetEvent(s->hEventRead);
    SetEvent(s->hEventWrite);
  }

  ReleaseMutex(s->hMutex);
}

#if defined (IN_FFSOX_BS1770) // {
void samples_reset_lufs(samples_t *s)
{
  //DMARKLN();
#if 1
  s->adaptive=0==memcmp(s->t->fn->utf8,"http://",strlen("http://"))
      ||0==memcmp(s->t->fn->utf8,"https://",strlen("https://"));
#else
  s->adaptive=1;
#endif
  bs1770_lufs_reset(&s->lufs,s->rate,s->nch,s->adaptive);

  if (0!=s->adaptive) {
    s->adaptive=1;
    s->q=DB2Q(DEFAULT_GAIN);
  }
}
#endif // }

void samples_clear(samples_t *s)
{
  //DMARKLN();
#if defined (IN_FFSOX_BS1770) // {
  samples_reset_lufs(s);
#endif // }
  s->in.wp=s->in.rep;
  s->time=0.0;
  s->state=STATE_RUN;
}

#if defined (IN_FFSOX_BS1770) // {
samples_t *samples_realloc(samples_t *s, track_t *t, size_t nsam,
    int nch, int bps, double rate)
#else // } {
samples_t *samples_realloc(samples_t *s, track_t *t, size_t nsam,
    int nch, int bps)
#endif // }
{
  enum { FACTOR=3 };
  size_t size;
  char *rep;

  //DMARKLN();
#if defined (FFSOX_DECODER_YASAPI) // [
  s->nsam=nsam;
#endif // ]
  size=nsam*nch*(bps/8);

  if (0==s->size) {
    if (NULL==(s->in.rep=malloc(FACTOR*size)))
      goto error;

    s->size=size;
  }
  else if (s->size<size) {
    if (NULL==(rep=realloc(s->in.rep,FACTOR*size)))
      goto error;

    s->in.rep=rep;
    s->size=size;
  }

  s->bps=bps;
#if defined (IN_FFSOX_BS1770) // {
  s->nch=nch;
  s->rate=rate;
#endif // }
  s->t=t;
  s->in.mp=s->in.rep+size;
  s->out.rep=s->in.rep+size;
  s->q=track_get_q(t);

  return s;
error:
  return NULL;
}

///////////////////////////////////////////////////////////////////////////////
void samples_wait_lock(samples_t *s)
{
  WaitForSingleObject(s->hMutex,INFINITE);

  while (STATE_RUN==s->state&&s->in.wp==s->in.mp) {
    SignalObjectAndWait(s->hMutex,s->hEventWrite,INFINITE,FALSE);
    WaitForSingleObject(s->hMutex,INFINITE);
  }
}

void samples_notify_unlock(samples_t *s)
{
  SetEvent(s->hEventRead);
  ReleaseMutex(s->hMutex);
}

///////////////////////////////////////////////////////////////////////////////
#define samples_scale_s16(s) \
  (16==(s)->bps?(s)->q:24==(s)->bps?(s)->q*(1<<8):(s)->q*(1<<16))
#define samples_scale_s32(s) \
  ((s)->q/(16==(s)->bps?1<<16:24==(s)->bps?1<<8:1))
#define samples_scale_f32(s) \
  ((s)->q*(16==(s)->bps?1<<15:24==(s)->bps?1<<23:1<<31))
#define samples_scale_f64(s) \
  ((s)->q*(16==(s)->bps?1<<15:24==(s)->bps?1<<23:1<<31))

///////////////////////////////////////////////////////////////////////////////
#define SAMPLES_CONVERT(x,wp,bytes,min,max) do { \
  double y; \
  int32_t z; \
 \
  y=round(q*x); \
  z=y<=min?min:y>=max?max:y; \
  memcpy(wp,&z,bytes); \
  wp+=bytes; \
} while (0)

///////////////////////////////////////////////////////////////////////////////
#if defined (IN_FFSOX_BS1770) // {
#define SAMPLES_COPY_I(id1,id2) \
void samples_copy_##id1##i(samples_t *s, uint8_t *rp, size_t nch, size_t n) \
{ \
  char *mwp; \
  char *wp; \
  id1##_t x; \
  size_t ch; \
  double q=samples_scale_##id1(s); \
  uint8_t *mrp=rp+nch*n*sizeof x; \
  size_t bytes=s->bps>>3; \
  int32_t min=(~0)<<(s->bps-1); \
  int32_t max=(~0)^(min); \
  bs1770_sample_##id2##_t sample; \
 \
  while (STATE_RUN==s->state&&rp<mrp) { \
    samples_wait_lock(s); \
    wp=(char *)s->in.wp; \
    mwp=s->in.mp; \
 \
    while (STATE_KILL!=s->state&&rp<mrp&&wp<mwp) { \
      for (ch=0;ch<nch;++ch) { \
        memcpy(&x,rp,sizeof x); \
        rp+=sizeof x; \
        SAMPLES_CONVERT(x,wp,bytes,min,max); \
        sample[ch]=x; \
      } \
 \
      s->lufs.vmt->add_sample_##id2(&s->lufs,sample); \
    } \
 \
    s->in.wp=wp; \
    samples_notify_unlock(s); \
  } \
}

SAMPLES_COPY_I(s16,i16)
SAMPLES_COPY_I(s32,i32)
SAMPLES_COPY_I(f32,f32)
SAMPLES_COPY_I(f64,f64)
#else // } {
#define SAMPLES_COPY_I(id) \
void samples_copy_##id##i(samples_t *s, uint8_t *rp, size_t nch, size_t n) \
{ \
  char *mwp; \
  char *wp; \
  id##_t x; \
  double q=samples_scale_##id(s); \
  uint8_t *mrp=rp+nch*n*sizeof x; \
  size_t bytes=s->bps>>3; \
  int32_t min=(~0)<<(s->bps-1); \
  int32_t max=(~0)^(min); \
 \
  while (STATE_RUN==s->state&&rp<mrp) { \
    samples_wait_lock(s); \
    wp=(char *)s->in.wp; \
    mwp=s->in.mp; \
 \
    while (STATE_KILL!=s->state&&rp<mrp&&wp<mwp) { \
      memcpy(&x,rp,sizeof x); \
      rp+=sizeof x; \
      SAMPLES_CONVERT(x,wp,bytes,min,max); \
    } \
 \
    s->in.wp=wp; \
    samples_notify_unlock(s); \
  } \
}

SAMPLES_COPY_I(s16)
SAMPLES_COPY_I(s32)
SAMPLES_COPY_I(f32)
SAMPLES_COPY_I(f64)
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (IN_FFSOX_BS1770) // {
#define SAMPLES_COPY_P(id1,id2) \
void samples_copy_##id1##p(samples_t *s, uint8_t **rp, size_t nch, size_t n) \
{ \
  char *mwp; \
  char *wp; \
  id1##_t x; \
  size_t i=0,ch; \
  double q=samples_scale_##id1(s); \
  size_t bytes=s->bps>>3; \
  int32_t min=(~0)<<(s->bps-1); \
  int32_t max=(~0)^(min); \
  bs1770_sample_##id2##_t sample; \
 \
  while (STATE_RUN==s->state&&i<n) { \
    samples_wait_lock(s); \
    wp=(char *)s->in.wp; \
    mwp=s->in.mp; \
 \
    while (STATE_KILL!=s->state&&i<n&&wp<mwp) { \
      for (ch=0;ch<nch;++ch) { \
        memcpy(&x,rp[ch]+i*sizeof x,sizeof x); \
        SAMPLES_CONVERT(x,wp,bytes,min,max); \
        sample[ch]=x; \
      } \
 \
      s->lufs.vmt->add_sample_##id2(&s->lufs,sample); \
      ++i; \
    } \
 \
    s->in.wp=wp; \
    samples_notify_unlock(s); \
  } \
}

SAMPLES_COPY_P(s16,i16)
SAMPLES_COPY_P(s32,i16)
SAMPLES_COPY_P(f32,f32)
SAMPLES_COPY_P(f64,f32)
#else // } {
#define SAMPLES_COPY_P(id) \
void samples_copy_##id##p(samples_t *s, uint8_t **rp, size_t nch, size_t n) \
{ \
  char *mwp; \
  char *wp; \
  id##_t x; \
  size_t i=0,ch; \
  double q=samples_scale_##id(s); \
  size_t bytes=s->bps>>3; \
  int32_t min=(~0)<<(s->bps-1); \
  int32_t max=(~0)^(min); \
 \
  while (STATE_RUN==s->state&&i<n) { \
    samples_wait_lock(s); \
    wp=(char *)s->in.wp; \
    mwp=s->in.mp; \
 \
    while (STATE_KILL!=s->state&&i<n&&wp<mwp) { \
      for (ch=0;ch<nch;++ch) { \
        memcpy(&x,rp[ch]+i*sizeof x,sizeof x); \
        SAMPLES_CONVERT(x,wp,bytes,min,max); \
      } \
 \
      ++i; \
    } \
 \
    s->in.wp=wp; \
    samples_notify_unlock(s); \
  } \
}

SAMPLES_COPY_P(s16)
SAMPLES_COPY_P(s32)
SAMPLES_COPY_P(f32)
SAMPLES_COPY_P(f64)
#endif // }

///////////////////////////////////////////////////////////////////////////////
#if defined (IN_FFSOX_BS1770) // {
void samples_adapt(samples_t *s, double *q)
{
  if (0<s->adaptive) {
    if (s->adaptive<DEFAULT_ADAPTIVE) {
      double weight=(double)s->adaptive/DEFAULT_ADAPTIVE;
      double db=s->lufs.vmt->get_cur_gain(&s->lufs,BS1770_R128_REFERENCE);

      *q=DB2Q((1.0-weight)*DEFAULT_GAIN+weight*db);
#if ! defined (DEBUG) // {
      ++s->adaptive;
#endif // }
    }
    else
      s->lufs.vmt->get_cur_gain_q(&s->lufs,BS1770_R128_REFERENCE,q);

#if defined (DEBUG) // {
    DPRINTF("REPLAY GAIN: %I64u, %.01f\n",s->adaptive,BS1770_Q2DB(*q));

    if (s->adaptive+1<INT_MAX)  
      ++s->adaptive;
#endif // }
  }
}
#endif // }

void samples_copy(samples_t *s, ffsox_ar_t *ar)
{
  AVFrame *frame=ffsox_ar_frame_internal(ar);
#if defined (FFSOX_DEPRECATED) // [
  int nch=frame->channels;
#else // ] [
  int nch=av_frame_get_channels(frame);
#endif // ]

  //DMARKLN();

  switch (frame->format) {
  // interleaved:
  case AV_SAMPLE_FMT_S16:
    samples_copy_s16i(s,frame->data[0],nch,frame->nb_samples);
    break;
  case AV_SAMPLE_FMT_S32:
    samples_copy_s32i(s,frame->data[0],nch,frame->nb_samples);
    break;
  case AV_SAMPLE_FMT_FLT:
    samples_copy_f32i(s,frame->data[0],nch,frame->nb_samples);
    break;
  case AV_SAMPLE_FMT_DBL:
    samples_copy_f64i(s,frame->data[0],nch,frame->nb_samples);
    break;
  // planar:
  case AV_SAMPLE_FMT_S16P:
    samples_copy_s16p(s,frame->data,nch,frame->nb_samples);
    break;
  case AV_SAMPLE_FMT_S32P:
    samples_copy_s32p(s,frame->data,nch,frame->nb_samples);
    break;
  case AV_SAMPLE_FMT_FLTP:
    samples_copy_f32p(s,frame->data,nch,frame->nb_samples);
    break;
  case AV_SAMPLE_FMT_DBLP:
    samples_copy_f64p(s,frame->data,nch,frame->nb_samples);
    break;
  // unsupported:
  default:
    DPRINTF("Error: Unsupported sample format: \"%s\"\n",
        av_get_sample_fmt_name(frame->format));
    break;
  }

  //avcodec_get_frame_defaults(frame);
}
