/*
 * timer.c
 * Copyright (C) 2010 - 2014 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

#if defined (NO_WAITABLE_TIMER) // {
// http://msdn.microsoft.com/en-us/library/windows/desktop/ms687003%28v=vs.85%29.aspx

timer_t *timer_init(timer_t *t)
{
  if (NULL==(t->hMutex=CreateMutex(NULL,FALSE,NULL)))
    goto fail;

  if (NULL==(t->hEvent=CreateEvent(NULL,FALSE,FALSE,NULL)))
    goto fail;

  if (NULL==(t->hTimerQueue=CreateTimerQueue()))
    goto fail;

  return t;
fail:
  timer_cleanup(t);

  return NULL;
}

void timer_cleanup(timer_t *t)
{
  if (NULL!=t->hTimerQueue)
    DeleteTimerQueue(t->hTimerQueue);

  if (NULL!=t->hEvent)
    CloseHandle(t->hEvent);

  if (NULL!=t->hMutex)
    CloseHandle(t->hMutex);
}

static VOID CALLBACK TimerRoutine(PVOID lpParam, BOOLEAN TimerOrWaitFired)
{
  timer_t *t=(timer_t *)lpParam;

  WaitForSingleObject(t->hMutex,INFINITE);

  if (STATE_KILL!=t->state)
    t->state=STATE_EOF;

  SetEvent(t->hEvent);
  ReleaseMutex(t->hMutex);
}

void timer_kill(timer_t *t)
{
  WaitForSingleObject(t->hMutex,INFINITE);
  t->state=STATE_KILL;
  SetEvent(t->hEvent);
  ReleaseMutex(t->hMutex);
}

void timer_sleep(timer_t *t, DWORD msInterval)
{
  HANDLE hTimer=NULL;
  BOOL state;

  WaitForSingleObject(t->hMutex,INFINITE);

  if (STATE_KILL!=t->state) {
    t->state=STATE_RUN;

    state=CreateTimerQueueTimer(
      &hTimer,				// _Out_     PHANDLE phNewTimer,
      t->hTimerQueue,		// _In_opt_  HANDLE TimerQueue,
      TimerRoutine,			// _In_      WAITORTIMERCALLBACK Callback,
      t,					// _In_opt_  PVOID Parameter,
      msInterval,			// _In_      DWORD DueTime,
      0,					// _In_      DWORD Period,
      WT_EXECUTEONLYONCE	// _In_      ULONG Flags
    );

	if (state) {
      while (STATE_RUN==t->state) {
        SignalObjectAndWait(t->hMutex,t->hEvent,INFINITE,FALSE);
        WaitForSingleObject(t->hMutex,INFINITE);
      }
    }
	else
	  t->state=STATE_KILL;
  }

  ReleaseMutex(t->hMutex);
}
#else // } {
// http://www.virtualdub.org/blog/pivot/entry.php?id=272

timer_t *timer_init(timer_t *t)
{
  t->hTimer=CreateWaitableTimer(
    NULL,           // _In_opt_  LPSECURITY_ATTRIBUTES lpTimerAttributes,
    FALSE,          // _In_      BOOL bManualReset,
    NULL            // _In_opt_  LPCTSTR lpTimerName
  );

  return NULL==t->hTimer?NULL:t;
}

void timer_cleanup(timer_t *t)
{
  if (NULL!=t->hTimer)
    CloseHandle(t->hTimer);
}

static void timer_set(timer_t *t, DWORD ms)
{
  LARGE_INTEGER ui;

  // http://msdn.microsoft.com/en-us/library/windows/desktop/ms687012%28v=vs.85%29.aspx
  // Positive values indicate absolute time. 
  // Negative values indicate relative time.
  // in 100 nanosecond intervals
  ui.QuadPart=-10000ll*ms;

  SetWaitableTimer(
    t->hTimer,    // _In_      HANDLE hTimer,
    &ui,          // _In_      const LARGE_INTEGER *pDueTime,
    0,            // _In_      LONG lPeriod,
    NULL,         // _In_opt_  PTIMERAPCROUTINE pfnCompletionRoutine,
    NULL,         // _In_opt_  LPVOID lpArgToCompletionRoutine,
    FALSE         // _In_      BOOL fResume
  );
}

void timer_kill(timer_t *t)
{
  t->state=STATE_KILL;
  CancelWaitableTimer(t->hTimer);
  timer_set(t,0);
}

void timer_sleep(timer_t *t, DWORD ms)
{
  if (STATE_KILL!=t->state) {
    t->state=STATE_RUN;
    timer_set(t,ms);
    WaitForSingleObject(t->hTimer,INFINITE);
  }
}
#endif // }
