/*
 * video_filter.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"
#include <libavfilter/buffersink.h>
#include <libavfilter/buffersrc.h>

// Adapted from init_filters() @ "<ffmpeg>/doc/examples/filering_video.c".
// Dont't forget to call avfilter_register_all()!!!
static int init_filters(video_filter_t *vf, AVFormatContext *pFormatContext,
    int vi, AVCodecContext *pCodecContext, const char *filters_descr)
{
    char args[512];
    int ret = 0;
    const AVFilter *buffersrc  = avfilter_get_by_name("buffer");
    const AVFilter *buffersink = avfilter_get_by_name("buffersink");
    AVFilterInOut *outputs = avfilter_inout_alloc();
    AVFilterInOut *inputs  = avfilter_inout_alloc();
    AVRational time_base = pFormatContext->streams[vi]->time_base;
    enum AVPixelFormat pix_fmts[] = { /*AV_PIX_FMT_GRAY8,*/ AV_PIX_FMT_NONE };

    vf->filter_graph = avfilter_graph_alloc();

    if (!outputs || !inputs || !vf->filter_graph) {
        ret = AVERROR(ENOMEM);
        goto end;
    }

    // buffer video source: the decoded frames from the decoder will be
    // inserted here.
    snprintf(args, sizeof(args),
            "video_size=%dx%d:pix_fmt=%d:time_base=%d/%d:pixel_aspect=%d/%d",
            pCodecContext->width,pCodecContext->height,pCodecContext->pix_fmt,
            time_base.num, time_base.den,
            pCodecContext->sample_aspect_ratio.num,
            pCodecContext->sample_aspect_ratio.den);

    ret = avfilter_graph_create_filter(&vf->buffersrc_ctx, buffersrc, "in",
                                       args, NULL, vf->filter_graph);

    if (ret < 0) {
        DVMESSAGE("avfilter_graph_create_filter() failed with %d (%s)",
            ret,av_err2str(ret));
        goto end;
    }

    // buffer video sink: to terminate the filter chain.
    ret = avfilter_graph_create_filter(&vf->buffersink_ctx, buffersink, "out",
                                       NULL, NULL, vf->filter_graph);
    if (ret < 0) {
        DVMESSAGE("avfilter_graph_create_filter() failed with %d (%s)",
            ret,av_err2str(ret));
        goto end;
    }

    ret = av_opt_set_int_list(vf->buffersink_ctx, "pix_fmts", pix_fmts,
                              AV_PIX_FMT_NONE, AV_OPT_SEARCH_CHILDREN);
    if (ret < 0) {
        DVMESSAGE("av_opt_set_int_list() failed with %d (%s)",
            ret,av_err2str(ret));
        goto end;
    }

    // Set the endpoints for the filter graph. The filter_graph will
    // be linked to the graph described by filters_descr.
    //
    // The buffer source output must be connected to the input pad of
    // the first filter described by filters_descr; since the first
    // filter input label is not specified, it is set to "in" by
    // default.
    outputs->name       = av_strdup("in");
    outputs->filter_ctx = vf->buffersrc_ctx;
    outputs->pad_idx    = 0;
    outputs->next       = NULL;

    // The buffer sink input must be connected to the output pad of
    // the last filter described by filters_descr; since the last
    // filter output label is not specified, it is set to "out" by
    // default.
    inputs->name       = av_strdup("out");
    inputs->filter_ctx = vf->buffersink_ctx;
    inputs->pad_idx    = 0;
    inputs->next       = NULL;

    if ((ret = avfilter_graph_parse_ptr(vf->filter_graph, filters_descr,
                                    &inputs, &outputs, NULL)) < 0)
        goto end;

    if ((ret = avfilter_graph_config(vf->filter_graph, NULL)) < 0)
        goto end;

end:
    avfilter_inout_free(&inputs);
    avfilter_inout_free(&outputs);

    return ret;
}

int video_filter_init(video_filter_t *vf, AVFormatContext *pFormatContext,
    int vi, AVCodecContext *pCodecContext, const char *filters_descr)
{
  int ret=0;

  vf->frame=NULL;
  vf->filter_graph=NULL;
  vf->buffersink_ctx=NULL;
  vf->buffersrc_ctx=NULL;

  if (!(vf->frame=av_frame_alloc())) {
    DMESSAGE("av_frame_alloc() failed");
    goto end;
  }

  ret=init_filters(vf,pFormatContext,vi,pCodecContext,filters_descr);

  if (ret<0) {
    DVMESSAGE("init_filters() failed with %d (%s)",ret,av_err2str(ret));
    goto end;
  }
end:
  return ret;
}

void video_filter_cleanup(video_filter_t *vf)
{
  vf->buffersrc_ctx=NULL;
  vf->buffersink_ctx=NULL;

  if (vf->filter_graph) {
    avfilter_graph_free(&vf->filter_graph);
    vf->filter_graph=NULL;
  }

  if (vf->frame) {
    av_free(vf->frame);
    vf->frame=NULL;
  }
}

int video_filter_push_frame(video_filter_t *vf, AVFrame *frame)
{
  return av_buffersrc_add_frame_flags(vf->buffersrc_ctx,frame,
      AV_BUFFERSRC_FLAG_KEEP_REF);
}

int video_filter_pull_frame(video_filter_t *vf)
{
  return av_buffersink_get_frame(vf->buffersink_ctx,vf->frame);
}
