/*
 * ffsox_audio_filter.c
 * Copyright (C) 2011, 2018 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "ffsox.h"

#if defined (FFSOX_FILTER_AR) // [
#define FFSOX_FILTER_PAN
///////////////////////////////////////////////////////////////////////////////
int ffsox_audio_filter_init(ffsox_audio_filter_t *af,
    AVFormatContext *pFormatContext, AVCodecContext *pCodecContext,
    const AVFrame *pFrame, int i, const char *filters_descr)
{
  AVRational time_base={ 1, pFrame->sample_rate };
#if defined (FFSOX_FILTER_PAN) // [
  const AVFilter  *abuffer,*volume,*pan,*sink;
  AVFilterContext *abuffer_ctx,*volume_ctx,*pan_ctx,*sink_ctx;
#else // ] [
  const AVFilter  *abuffer,*volume,*aformat,*sink;
  AVFilterContext *abuffer_ctx,*volume_ctx,*aformat_ctx,*sink_ctx;
#endif // ]
  int err;

  af->pFrame=NULL;
  af->pFilterGraph=NULL;

  ///////////////////////////////////////////////////////////////////////////
  af->pFrame=av_frame_alloc();
  if (af->pFrame) {
    DMESSAGE("av_frame_alloc() failed");
    err=AVERROR(ENOMEM);
    goto frame;
  }

  ///////////////////////////////////////////////////////////////////////////
  /* Create a new filtergraph, which will contain all the filters. */
  af->pFilterGraph=avfilter_graph_alloc();
  if (!af->pFilterGraph) {
    //fprintf(stderr, "Unable to create filter graph.\n");
    err=AVERROR(ENOMEM);
    goto filter_graph;
  }

  ///////////////////////////////////////////////////////////////////////////
  /* Create the abuffer filter;
   * it will be used for feeding the data into the graph. */
  abuffer=avfilter_get_by_name("abuffer");
  if (!abuffer) {
    //fprintf(stderr, "Could not find the abuffer filter.\n");
    err=AVERROR_FILTER_NOT_FOUND;
    goto abuffer;
  }

  abuffer_ctx=avfilter_graph_alloc_filter(af->pFilterGraph,abuffer,"src");
  if (!abuffer_ctx) {
    //fprintf(stderr,"Could not allocate the abuffer instance.\n");
    err=AVERROR(ENOMEM);
    goto abuffer_ctx;
  }

  /* Set the filter options through the AVOptions API. */
#if 0 // [
  av_get_channel_layout_string(ch_layout, sizeof(ch_layout), 0,
    pFrame->channel_layout);
  av_opt_set(abuffer_ctx,"channel_layout",
    ch_layout,AV_OPT_SEARCH_CHILDREN);
  av_opt_set(abuffer_ctx,"sample_fmt",
    av_get_sample_fmt_name(pFrame->format),AV_OPT_SEARCH_CHILDREN);
#else // ] [
  av_opt_set_int(abuffer_ctx,"channel_layout",
    pFrame->channel_layout,AV_OPT_SEARCH_CHILDREN);
  av_opt_set_int(abuffer_ctx,"sample_fmt",
    pFrame->format,AV_OPT_SEARCH_CHILDREN);
#endif // ]
  av_opt_set_q(abuffer_ctx,"time_base",
    time_base,AV_OPT_SEARCH_CHILDREN);
  av_opt_set_int(abuffer_ctx,"sample_rate",
    pFrame->sample_rate,AV_OPT_SEARCH_CHILDREN);

  /* Now initialize the filter; we pass NULL options, since we have already
   * set all the options above. */
  err=avfilter_init_str(abuffer_ctx,NULL);
  if (err < 0) {
    //fprintf(stderr, "Could not initialize the abuffer filter.\n");
    goto abuffer_ctx_init;
  }

  ///////////////////////////////////////////////////////////////////////////
  /* Create volume filter. */
  volume=avfilter_get_by_name("volume");
  if (!volume) {
    //fprintf(stderr, "Could not find the volume filter.\n");
    err=AVERROR_FILTER_NOT_FOUND;
    goto volume;
  }

  volume_ctx=avfilter_graph_alloc_filter(af->pFilterGraph,volume,"volume");
  if (!volume_ctx) {
    //fprintf(stderr, "Could not allocate the volume instance.\n");
    err=AVERROR(ENOMEM);
    goto volume_ctx;
  }

#if 0 // [
  /* A different way of passing the options is as key/value pairs in a
   * dictionary. */
  av_dict_set(&options_dict, "volume", AV_STRINGIFY(VOLUME_VAL), 0);
  err = avfilter_init_dict(volume_ctx, &options_dict);
  av_dict_free(&options_dict);
  if (err < 0) {
    fprintf(stderr, "Could not initialize the volume filter.\n");
    return err;
  }
#else // ] [
  av_opt_set_double(volume_ctx,"volume",0.9,AV_OPT_SEARCH_CHILDREN);

#if 0 // [
  err=avfilter_init_dict(volume_ctx,NULL);
#else // ] [
  err=avfilter_init_str(volume_ctx,NULL);
#endif // ]
  if (err < 0) {
    //fprintf(stderr, "Could not initialize the volume filter.\n");
    goto volume_ctx_init;
  }
#endif // ]

  ///////////////////////////////////////////////////////////////////////////
  /* Create the aformat filter;
   * it ensures that the output is of the format we want. */
#if defined (FFSOX_FILTER_PAN) // [
  pan=avfilter_get_by_name("pan");
  if (!pan) {
    //fprintf(stderr, "Could not find the pan filter.\n");
    err=AVERROR_FILTER_NOT_FOUND;
    goto pan;
  }

  pan_ctx=avfilter_graph_alloc_filter(af->pFilterGraph,pan,"pan");
  if (!pan_ctx) {
    //fprintf(stderr, "Could not allocate the pan instance.\n");
    err=AVERROR(ENOMEM);
    goto pan_ctx;
  }
#else // ] [
  aformat=avfilter_get_by_name("aformat");
  if (!aformat) {
    //fprintf(stderr, "Could not find the aformat filter.\n");
    err=AVERROR_FILTER_NOT_FOUND;
    goto aformat;
  }

  aformat_ctx=avfilter_graph_alloc_filter(af->pFilterGraph,aformat,"aformat");
  if (!aformat_ctx) {
    //fprintf(stderr, "Could not allocate the aformat instance.\n");
    err=AVERROR(ENOMEM);
    goto aformat_ctx;
  }
#endif // ]

#if defined (FFSOX_FILTER_PAN) // [
//#error pan
  err = avfilter_init_str(pan_ctx, NULL);
  if (err < 0) {
    //av_log(NULL, AV_LOG_ERROR, "Could not initialize the aformat filter.\n");
    goto pan_ctx_init;
  }
#else // ] [
#if 0 // [
  /* A third way of passing the options is in a string of the form
   * key1=value1:key2=value2.... */
  snprintf(options_str, sizeof(options_str),
         "sample_fmts=%s:sample_rates=%d:channel_layouts=0x%"PRIx64,
         av_get_sample_fmt_name(AV_SAMPLE_FMT_S16), 44100,
         (uint64_t)AV_CH_LAYOUT_STEREO);
  err = avfilter_init_str(aformat_ctx, options_str);
  if (err < 0) {
    av_log(NULL, AV_LOG_ERROR, "Could not initialize the aformat filter.\n");
    return err;
  }
#else // ] [
  #error avformat
#endif // ]
#endif // ]

  /* Finally create the abuffersink filter;
   * it will be used to get the filtered data out of the graph. */
  sink=avfilter_get_by_name("sink");
  if (!sink) {
    //fprintf(stderr, "Could not find the abuffersink filter.\n");
    err=AVERROR_FILTER_NOT_FOUND;
    goto sink;
  }

  sink_ctx=avfilter_graph_alloc_filter(af->pFilterGraph,sink,"sink");
  if (!sink_ctx) {
    //fprintf(stderr, "Could not allocate the abuffersink instance.\n");
    err=AVERROR(ENOMEM);
    goto sink_ctx;
  }

  /* This filter takes no options. */
  err=avfilter_init_str(sink_ctx, NULL);
  if (err<0) {
    //fprintf(stderr, "Could not initialize the abuffersink instance.\n");
    //return err;
    goto sink_ctx_init;
  }

  /* Connect the filters;
   * in this simple case the filters just form a linear chain. */
  err=avfilter_link(abuffer_ctx,0,volume_ctx,0);
#if defined (FFSOX_FILTER_PAN) // [
  if (0<=err)
    err=avfilter_link(volume_ctx,0,pan_ctx,0);
  if (0<=err)
    err=avfilter_link(pan_ctx,0,sink_ctx,0);
#else // ] [
  if (0<=err)
    err=avfilter_link(volume_ctx,0,aformat_ctx,0);
  if (0<=err)
    err=avfilter_link(aformat_ctx,0,sink_ctx,0);
#endif // ]
  if (err<0) {
    //fprintf(stderr, "Error connecting filters\n");
    goto link;
  }

  /* Configure the graph. */
  err=avfilter_graph_config(af->pFilterGraph,NULL);
  if (err<0) {
    //av_log(NULL,AV_LOG_ERROR,"Error configuring the filter graph\n");
    goto config;
  }

#if 0 // [
  *graph = filter_graph;
  *src   = abuffer_ctx;
  *sink  = abuffersink_ctx;
#endif // ]
#if 0 // [
  if (ffsox_audio_filter_init_filters(af,pFormatContext,i,pCodecContext,
      pFrame,filters_descr)<0) {
    DVMESSAGE("init_filters() failed with %d (%s)",ret,av_err2str(ret));
    goto filters;
  }
#endif // ]

  return 0;
//cleanup:
config:
link:
sink_ctx_init:
sink_ctx:
sink:
#if defined (FFSOX_FILTER_PAN) // [
pan_ctx_init:
pan_ctx:
pan:
#else // ] [
aformat_ctx:
aformat:
#endif // ]
volume_ctx_init:
volume_ctx:
volume:
abuffer_ctx_init:
abuffer_ctx:
abuffer:
  avfilter_graph_free(&af->pFilterGraph);
  af->pFilterGraph=NULL;
filter_graph:
  av_free(af->pFrame);
  af->pFrame=NULL;
frame:
  return err;
}
#endif // ]
