/*
 * ffsox_filter.c
 * Copyright (C) 2011, 2018 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "ffsox.h"

///////////////////////////////////////////////////////////////////////////////
static int ffsox_filter_init_filters(ffsox_filter_t *f,
    AVFormatContext *pFormatContext, int i, AVCodecContext *pCodecContext,
    const AVFrame *pFrame, const char *filters_descr)
{
    AVRational time_base={ 1, pFrame->sample_rate };
    const AVFilter  *abuffer,*volume,*aformat;
    AVFilterContext *abuffer_ctx,*volume_ctx,*aformat_ctx;
#if 0 // [
    char ch_layout[64];
#endif // ]
    int err;

    ///////////////////////////////////////////////////////////////////////////
    /* Create a new filtergraph, which will contain all the filters. */
    f->filter_graph = avfilter_graph_alloc();
    if (!f->filter_graph) {
        //fprintf(stderr, "Unable to create filter graph.\n");
        err=AVERROR(ENOMEM);
        goto filter_graph;
    }

    ///////////////////////////////////////////////////////////////////////////
    /* Create the abuffer filter;
     * it will be used for feeding the data into the graph. */
    abuffer = avfilter_get_by_name("abuffer");
    if (!abuffer) {
        //fprintf(stderr, "Could not find the abuffer filter.\n");
        err=AVERROR_FILTER_NOT_FOUND;
        goto abuffer;
    }

    abuffer_ctx = avfilter_graph_alloc_filter(f->filter_graph, abuffer, "src");
    if (!abuffer_ctx) {
        //fprintf(stderr, "Could not allocate the abuffer instance.\n");
        err=AVERROR(ENOMEM);
        goto abuffer_ctx;
    }

    /* Set the filter options through the AVOptions API. */
#if 0 // [
    av_get_channel_layout_string(ch_layout, sizeof(ch_layout), 0,
        pFrame->channel_layout);
    av_opt_set(abuffer_ctx,"channel_layout",
        ch_layout,AV_OPT_SEARCH_CHILDREN);
    av_opt_set(abuffer_ctx,"sample_fmt",
        av_get_sample_fmt_name(pFrame->format),AV_OPT_SEARCH_CHILDREN);
#else // ] [
    av_opt_set_int(abuffer_ctx,"channel_layout",
        pFrame->channel_layout,AV_OPT_SEARCH_CHILDREN);
    av_opt_set_int(abuffer_ctx,"sample_fmt",
        pFrame->format,AV_OPT_SEARCH_CHILDREN);
#endif // ]
    av_opt_set_q(abuffer_ctx,"time_base",
        time_base,AV_OPT_SEARCH_CHILDREN);
    av_opt_set_int(abuffer_ctx,"sample_rate",
        pFrame->sample_rate,AV_OPT_SEARCH_CHILDREN);

    /* Now initialize the filter; we pass NULL options, since we have already
     * set all the options above. */
    err = avfilter_init_str(abuffer_ctx, NULL);
    if (err < 0) {
        //fprintf(stderr, "Could not initialize the abuffer filter.\n");
        goto abuffer_ctx_init;
    }

    ///////////////////////////////////////////////////////////////////////////
    /* Create volume filter. */
    volume = avfilter_get_by_name("volume");
    if (!volume) {
        //fprintf(stderr, "Could not find the volume filter.\n");
        err=AVERROR_FILTER_NOT_FOUND;
        goto volume;
    }

    volume_ctx=avfilter_graph_alloc_filter(f->filter_graph, volume, "volume");
    if (!volume_ctx) {
        //fprintf(stderr, "Could not allocate the volume instance.\n");
        err=AVERROR(ENOMEM);
        goto volume_ctx;
    }

#if 0 // [
    /* A different way of passing the options is as key/value pairs in a
     * dictionary. */
    av_dict_set(&options_dict, "volume", AV_STRINGIFY(VOLUME_VAL), 0);
    err = avfilter_init_dict(volume_ctx, &options_dict);
    av_dict_free(&options_dict);
    if (err < 0) {
        fprintf(stderr, "Could not initialize the volume filter.\n");
        return err;
    }
#else // ] [
    av_opt_set_double(volume_ctx,"volume",0.9,AV_OPT_SEARCH_CHILDREN);

    err = avfilter_init_dict(volume_ctx, NULL);
    if (err < 0) {
        //fprintf(stderr, "Could not initialize the volume filter.\n");
        goto volume_ctx_init;
    }
#endif // ]

    ///////////////////////////////////////////////////////////////////////////
    /* Create the aformat filter;
     * it ensures that the output is of the format we want. */
    aformat = avfilter_get_by_name("aformat");
    if (!aformat) {
        //fprintf(stderr, "Could not find the aformat filter.\n");
        err=AVERROR_FILTER_NOT_FOUND;
        goto aformat;
    }

    aformat_ctx=avfilter_graph_alloc_filter(f->filter_graph,aformat,"aformat");
    if (!aformat_ctx) {
        //fprintf(stderr, "Could not allocate the aformat instance.\n");
        err=AVERROR(ENOMEM);
        goto aformat_ctx;
    }

#if 0 // [
    /* A third way of passing the options is in a string of the form
     * key1=value1:key2=value2.... */
    snprintf(options_str, sizeof(options_str),
             "sample_fmts=%s:sample_rates=%d:channel_layouts=0x%"PRIx64,
             av_get_sample_fmt_name(AV_SAMPLE_FMT_S16), 44100,
             (uint64_t)AV_CH_LAYOUT_STEREO);
    err = avfilter_init_str(aformat_ctx, options_str);
    if (err < 0) {
        av_log(NULL, AV_LOG_ERROR, "Could not initialize the aformat filter.\n");
        return err;
    }
#else // ] [
#endif // ]

#if 0 // [
    /* Finally create the abuffersink filter;
     * it will be used to get the filtered data out of the graph. */
    abuffersink = avfilter_get_by_name("abuffersink");
    if (!abuffersink) {
        fprintf(stderr, "Could not find the abuffersink filter.\n");
        return AVERROR_FILTER_NOT_FOUND;
    }

    abuffersink_ctx = avfilter_graph_alloc_filter(filter_graph, abuffersink, "sink");
    if (!abuffersink_ctx) {
        fprintf(stderr, "Could not allocate the abuffersink instance.\n");
        return AVERROR(ENOMEM);
    }

    /* This filter takes no options. */
    err = avfilter_init_str(abuffersink_ctx, NULL);
    if (err < 0) {
        fprintf(stderr, "Could not initialize the abuffersink instance.\n");
        return err;
    }

    /* Connect the filters;
     * in this simple case the filters just form a linear chain. */
    err = avfilter_link(abuffer_ctx, 0, volume_ctx, 0);
    if (err >= 0)
        err = avfilter_link(volume_ctx, 0, aformat_ctx, 0);
    if (err >= 0)
        err = avfilter_link(aformat_ctx, 0, abuffersink_ctx, 0);
    if (err < 0) {
        fprintf(stderr, "Error connecting filters\n");
        return err;
    }

    /* Configure the graph. */
    err = avfilter_graph_config(filter_graph, NULL);
    if (err < 0) {
        av_log(NULL, AV_LOG_ERROR, "Error configuring the filter graph\n");
        return err;
    }

    *graph = filter_graph;
    *src   = abuffer_ctx;
    *sink  = abuffersink_ctx;
#endif // ]
    return 0;
//cleanup:
aformat_ctx:
aformat:
volume_ctx_init:
volume_ctx:
volume:
abuffer_ctx_init:
abuffer_ctx:
abuffer:
    avfilter_graph_free(&f->filter_graph);
filter_graph:
    return err;
}

int ffsox_filter_init(ffsox_filter_t *f, AVFormatContext *pFormatContext,
    AVCodecContext *pCodecContext ,const AVFrame *pFrame, int i,
    const char *filters_descr)
{
  f->frame=NULL;
  f->filter_graph=NULL;
#if 0 // [
  f->buffersink_ctx=NULL;
  f->buffersrc_ctx=NULL;
#else // ] [
  //f->abuffer_ctx=NULL;
#endif // ]

  if (!(f->frame=av_frame_alloc())) {
    DMESSAGE("av_frame_alloc() failed");
    goto frame;
  }

  if (ffsox_filter_init_filters(f,pFormatContext,i,pCodecContext,pFrame,
      filters_descr)<0) {
    DVMESSAGE("init_filters() failed with %d (%s)",ret,av_err2str(ret));
    goto filters;
  }

  return 0;
filters:
  av_free(f->frame);
  f->frame=NULL;
frame:
  return -1;
}
