/*
 * ffsox_lame.c
 * Copyright (C) 2011, 2012 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include "ffsox.h"

// http://permalink.gmane.org/gmane.comp.audio.mp3.lame/4005

#if ! defined(WIN32)
  #define stricmp strcasecmp
  #define strnicmp strncasecmp
#endif

///////////////////////////////////////////////////////////////////////////////
typedef struct priv {
  size_t audio_pos;
  ffsox_buf_t mp3;
  ffsox_buf_t pcm_l;
  ffsox_buf_t pcm_r;
  FILE *f;
  ffi_t ffi;
  ffsox_ar_t ar;
  lame_t l;
} priv_t;

///////////////////////////////////////////////////////////////////////////////
#if defined (WIN32)
static wchar_t* utf8ToUtf16(char const* mbstr)
{
  // additional Byte-Order-Marker
  int n=MultiByteToWideChar(CP_UTF8, 0, mbstr, -1, NULL, 0);
  wchar_t *wstr=malloc((n+1)*sizeof(wstr[0]));

  if (NULL!=wstr) {
    wstr[0] = 0xfeff; // BOM
    n=MultiByteToWideChar(CP_UTF8,0,mbstr,-1,wstr+1,n);

    if (0==n) {
      free(wstr);
      wstr=NULL;
    }
  }

  return wstr;
}

static void priv_set_textinfo(priv_t *p,char *key,char *value)
{
  wchar_t *wvalue=NULL;

  if (NULL==(wvalue=utf8ToUtf16(value)))
	goto cleanup;

  id3tag_set_textinfo_utf16(p->l,key,wvalue);
cleanup:
  if (NULL!=wvalue)
    free(wvalue);
}
#endif

static void priv_write_id3tags(priv_t *p)
{
  AVFormatContext *ic=p->ffi.ic;
  AVDictionaryEntry *mtag=NULL;

  while (NULL!=(mtag=av_dict_get(ic->metadata,"",mtag,
      AV_DICT_IGNORE_SUFFIX))) {
#if defined (WIN32)
    if (!stricmp("artist",mtag->key))
      priv_set_textinfo(p,"TPE1",mtag->value);
    else if (!stricmp("album",mtag->key))
      priv_set_textinfo(p,"TALB",mtag->value);
    else if (!stricmp("year",mtag->key)||!stricmp("date",mtag->key))
      priv_set_textinfo(p,"TYER",mtag->value);
    else if (!stricmp("track",mtag->key))
      priv_set_textinfo(p,"TRCK",mtag->value);
    else if (!stricmp("title",mtag->key))
      priv_set_textinfo(p,"TIT2",mtag->value);
    else if (!stricmp("genre",mtag->key))
      priv_set_textinfo(p,"TCON",mtag->value);
#else
    if (!stricmp("artist",mtag->key))
      id3tag_set_artist(p->l,mtag->value);
    else if (!stricmp("album",mtag->key))
      id3tag_set_album(p->l,mtag->value);
    else if (!stricmp("year",mtag->key)||!stricmp("date",mtag->key))
      id3tag_set_year(p->l,mtag->value);
    else if (!stricmp("track",mtag->key))
      id3tag_set_track(p->l,mtag->value);
    else if (!stricmp("title",mtag->key))
      id3tag_set_title(p->l,mtag->value);
    else if (!stricmp("genre",mtag->key))
      id3tag_set_genre(p->l,mtag->value);
#endif
  }
}

///////////////////////////////////////////////////////////////////////////////
static priv_t *priv_cleanup(priv_t *p)
{
  if (NULL!=p->l)
    lame_close(p->l);

  if (NULL!=p->f)
    fclose(p->f);

  ffsox_ar_cleanup(&p->ar);
  ffi_cleanup(&p->ffi);
  ffsox_buf_cleanup(&p->pcm_r);
  ffsox_buf_cleanup(&p->pcm_l);
  ffsox_buf_cleanup(&p->mp3);
}

static priv_t *priv_init(priv_t *p, pb_str_t *fn1, pb_str_t *fn2, int mode,
    double q, double quality)
{
  AVCodecContext *dc;

  p->audio_pos=0;

  if (NULL==ffsox_buf_init(&p->mp3,4096))
    return NULL;

  if (NULL==ffsox_buf_init(&p->pcm_l,0))
    return NULL;

  if (NULL==ffsox_buf_init(&p->pcm_r,0))
    return NULL;

  if (NULL==ffi_init(&p->ffi,fn1,mode|FFI_STREAM_META|FFI_OPEN|FFI_STEREO))
    return NULL;

  if (NULL==ffsox_ar_init_ffi(&p->ar,q,&p->ffi))
    return NULL;

#if defined (WIN32)
  if (NULL==(p->f=_tfopen(PS_TCS(fn2),_T("wb"))))
    return NULL;
#else
  if (NULL==(p->f=fopen(PS_UTF8(fn2),"wb")))
    return NULL;
#endif

  if (NULL==(p->l=lame_init()))
    return NULL;

  // input.
  dc=p->ffi.ast->dc;
  lame_set_in_samplerate(p->l,dc->sample_rate);
  lame_set_num_channels(p->l,dc->channels);

  // encoder
  if (vbr_off==lame_get_VBR(p->l))
    lame_set_VBR(p->l,vbr_default);

  lame_set_VBR_quality(p->l,quality);

  // id3tag
  id3tag_init(p->l);
  id3tag_add_v2(p->l);
  priv_write_id3tags(p);

  // initialiazation done
  lame_set_write_id3tag_automatic(p->l,0);
  lame_init_params(p->l);

  return p;
}

static void priv_head(priv_t *p)
{
  size_t size;

  size=lame_get_id3v2_tag(p->l,p->mp3.wp,0);
  ffsox_buf_realloc(&p->mp3,size);
  size=lame_get_id3v2_tag(p->l,p->mp3.wp,ffsox_buf_avail(&p->mp3));
  fwrite(p->mp3.wp,1,size,p->f);

  fflush(p->f);
  p->audio_pos=ftell(p->f);
}

static void priv_tail(priv_t *p)
{
  size_t size;

  // flush encoder.
  ffsox_buf_realloc(&p->mp3,7200);
  size=lame_encode_flush(p->l,p->mp3.wp,ffsox_buf_avail(&p->mp3));
  fwrite(p->mp3.wp,1,size,p->f);

  // write id3v1.
  size=lame_get_id3v1_tag(p->l,p->mp3.wp,0);
  ffsox_buf_realloc(&p->mp3,size);
  size=lame_get_id3v1_tag(p->l,p->mp3.wp,ffsox_buf_avail(&p->mp3));
  fwrite(p->mp3.wp,1,size,p->f);

  fflush(p->f);
  fseek(p->f,p->audio_pos,SEEK_SET);

  // write XING header.
  size=lame_get_lametag_frame(p->l,p->mp3.wp,0);
  ffsox_buf_realloc(&p->mp3,size);
  size=lame_get_lametag_frame(p->l,p->mp3.wp,ffsox_buf_avail(&p->mp3));
  fwrite(p->mp3.wp,1,size,p->f);
}

///////////////////////////////////////////////////////////////////////////////
#define scale_s16(q) \
  (-(q)/INT16_MIN)
#define scale_s32(q) \
  (-(q)/INT32_MIN)
#define scale_f32(q) \
  (q)
#define scale_f64(q) \
  (q)

///////////////////////////////////////////////////////////////////////////////
#define PRIV_ENCODE_I(id) \
static int priv_encode_##id##i(priv_t *p, AVFrame *frame) \
{ \
  enum { NCH=2 }; \
  ffsox_##id##_t x; \
  int nb_samples=frame->nb_samples; \
  uint8_t *rp=frame->data[0]; \
  uint8_t *mrp=rp+NCH*nb_samples*sizeof x; \
  double *pcm_l=(double *)p->pcm_l.wp,*wpl=pcm_l; \
  double *pcm_r=(double *)p->pcm_r.wp,*wpr=pcm_r; \
  double q=scale_##id(p->ar.q); \
 \
  while (rp<mrp) { \
	memcpy(&x,rp,sizeof x); \
	rp+=sizeof x; \
    *wpl++=q*x; \
 \
	memcpy(&x,rp,sizeof x); \
	rp+=sizeof x; \
    *wpr++=q*x; \
  } \
 \
  return lame_encode_buffer_ieee_double(p->l,pcm_l,pcm_r,nb_samples, \
      p->mp3.wp,ffsox_buf_avail(&p->mp3)); \
}

PRIV_ENCODE_I(s16)
PRIV_ENCODE_I(s32)
PRIV_ENCODE_I(f32)
PRIV_ENCODE_I(f64)

///////////////////////////////////////////////////////////////////////////////
#define PRIV_ENCODE_P(id) \
static int priv_encode_##id##p(priv_t *p, AVFrame *frame) \
{ \
  ffsox_##id##_t x; \
  int nb_samples=frame->nb_samples; \
  uint8_t *rpl=frame->data[0]; \
  uint8_t *rpr=frame->data[1]; \
  uint8_t *mrpl=rpl+nb_samples*sizeof x; \
  double *pcm_l=(double *)p->pcm_l.wp,*wpl=pcm_l; \
  double *pcm_r=(double *)p->pcm_r.wp,*wpr=pcm_r; \
  double q=scale_##id(p->ar.q); \
 \
  while (rpl<mrpl) { \
	memcpy(&x,rpl,sizeof x); \
	rpl+=sizeof x; \
    *wpl++=q*x; \
 \
	memcpy(&x,rpr,sizeof x); \
	rpr+=sizeof x; \
    *wpr++=q*x; \
  } \
 \
  return lame_encode_buffer_ieee_double(p->l,pcm_l,pcm_r,nb_samples, \
      p->mp3.wp,ffsox_buf_avail(&p->mp3)); \
}

PRIV_ENCODE_P(s16)
PRIV_ENCODE_P(s32)
PRIV_ENCODE_P(f32)
PRIV_ENCODE_P(f64)

///////////////////////////////////////////////////////////////////////////////
static void priv_encode(priv_t *p)
{
  AVFrame *frame=p->ar.frame;
  int nb_samples=frame->nb_samples;
  size_t mp3_size=1.25*nb_samples+7200;
  size_t pcm_size=nb_samples*sizeof (double);
  int size;

  ffsox_buf_realloc(&p->mp3,mp3_size);
  ffsox_buf_realloc(&p->pcm_l,nb_samples*sizeof (double));
  ffsox_buf_realloc(&p->pcm_r,nb_samples*sizeof (double));

  switch (frame->format) {
  // interleaved.
  case AV_SAMPLE_FMT_S16:
    size=priv_encode_s16i(p,frame);
    break;
  case AV_SAMPLE_FMT_S32:
    size=priv_encode_s32i(p,frame);
    break;
  case AV_SAMPLE_FMT_FLT:
    size=priv_encode_f32i(p,frame);
    break;
  case AV_SAMPLE_FMT_DBL:
    size=priv_encode_f64i(p,frame);
    break;
  // planar.
  case AV_SAMPLE_FMT_S16P:
    size=priv_encode_s16p(p,frame);
    break;
  case AV_SAMPLE_FMT_S32P:
    size=priv_encode_s32p(p,frame);
    break;
  case AV_SAMPLE_FMT_FLTP:
    size=priv_encode_f32p(p,frame);
    break;
  case AV_SAMPLE_FMT_DBLP:
    size=priv_encode_f64p(p,frame);
    break;
  // not supported.
  default:
    size=0;
    break;
  }

  if (0<size)
    fwrite(p->mp3.wp,1,size,p->f);
}

///////////////////////////////////////////////////////////////////////////////
int ffsox_lame(pb_str_t *fn1, pb_str_t *fn2, int mode, double gain,
    double quality)
{
  int state=-1;
  priv_t p;

  memset(&p,0,sizeof p);

  if (NULL==priv_init(&p,fn1,fn2,mode|FFI_STEREO,gain,quality))
    goto priv_cleanup;

  priv_head(&p);
  ffsox_ar_next(&p.ar);

  while (ffsox_ar_valid(&p.ar)) {
    priv_encode(&p);
    ffsox_ar_next(&p.ar);
  }

  priv_tail(&p);
  state=0;
//cleanup:
  priv_cleanup(&p);
priv_cleanup:

  return state;
}
